from unittest import TestCase

import Orange

import orangecontrib.conformal as cp


class TestReadme(TestCase):
    def setUp(self):
        self.tab = Orange.data.Table('iris')
        self.train, self.test = next(cp.evaluation.RandomSampler(self.tab, 2, 1))
        self.train, self.calibrate = next(cp.evaluation.RandomSampler(self.train, 2, 1))

    def test_readme(self):
        nc = cp.nonconformity.InverseProbability(Orange.classification.NaiveBayesLearner())
        ic = cp.classification.InductiveClassifier(nc)
        r = cp.evaluation.run(ic, 0.1, cp.evaluation.RandomSampler(self.tab, 2, 1))
        self.assertGreater(r.accuracy(), 0.8)



"""Regression module contains methods for conformal regression.

Conformal regressors predict a range of values (not always a single value) under a given
significance level (error rate). Every regressors works in combination with a nonconformity measure
and on average predicts the correct value with the given error rate. Lower error rates result in
narrower ranges of predicted values.

Structure:

- ConformalRegressor
    - Inductive (:py:class:`InductiveRegressor`)
    - Cross (:py:class:`CrossRegressor`)
"""

from copy import deepcopy

import numpy as np
from sklearn.model_selection import KFold

from Orange.data import Instance, Unknown

from orangecontrib.conformal.base import ConformalPredictor
from orangecontrib.conformal.nonconformity import RegrNC


class PredictionRegr:
    """Conformal regression prediction object,
    which is produced by the :py:func:`ConformalRegressor.predict` method.

    Attributes:
        lo (float): Lowest value of the predicted range.
        hi (float): Highest value of the predicted range.

    Examples:
        >>> train, test = next(LOOSampler(Table('housing')))
        >>> ccr = CrossRegressor(AbsError(LinearRegressionLearner()), 5, train)
        >>> prediction = ccr.predict(test[0], 0.1)
        >>> print(prediction.width())
    """

    def __init__(self, lo, hi):
        """Initialize the prediction.

        Args:
            lo (float): Lowest value of the predicted range.
            hi (float): Highest value of the predicted range.
        """
        self.lo = lo
        self.hi = hi

    def range(self):
        """Predicted range: :py:attr:`lo`, :py:attr:`hi`."""
        return self.lo, self.hi

    def verdict(self, ref):
        """Conformal regression prediction is correct when the actual value appears
        in the predicted range.

        Args:
            ref: Reference/actual value

        Returns:
            True if the prediction is correct.
        """
        return self.lo <= ref <= self.hi

    def width(self):
        """Width of the predicted range: :py:attr:`hi` - :py:attr:`lo`."""
        if np.isnan(self.lo) or np.isnan(self.hi):
            return 0
        else:
            return self.hi-self.lo


class ConformalRegressor(ConformalPredictor):
    """Base class for conformal regression."""

    def __init__(self, nc_measure):
        """Verify that the nonconformity measure can be used for regression."""
        assert isinstance(nc_measure, RegrNC), "Inappropriate nonconformity measure for regression"
        self.nc_measure = nc_measure

    def __str__(self):
        return "{} ({})".format(self.__class__.__name__, self.nc_measure)

    def predict(self, example, eps):
        """Compute a regression prediction object for a given example and significance level.

        Function determines what is the :py:attr:`eps`-th lowest nonconformity score and computes
        the range of values that would result in a lower or equal nonconformity. This inverse
        of the nonconformity score is computed by the nonconformity measure's
        :py:func:`cp.nonconformity.RegrNC.predict` function.

        Args:
            example (Instance): Orange row instance.
            eps (float): Default significance level (error rate).

        Returns:
            PredictionRegr: Regression prediction object.
        """
        s = int(eps*(len(self.alpha)+1)) - 1
        s = min(max(s, 0), len(self.alpha)-1)
        nc = self.alpha[s]
        lo, hi = self.nc_measure.predict(example, nc)
        return PredictionRegr(lo, hi)

    def __call__(self, example, eps):
        """Compute predicted range for a given example and significance level.

        Args:
            example (Instance): Orange row instance.
            eps (float): Significance level (error rate).

        Returns:
            Predicted range as a pair (`PredictionRegr.lo`, `PredictionRegr.hi`)
        """
        pred = self.predict(example, eps)
        return pred.range()


class TransductiveRegressor(ConformalRegressor):
    """Transductive regression. TODO
    """
    pass


class InductiveRegressor(ConformalRegressor):
    """Inductive regression.

    Attributes:
        alpha: Nonconformity scores of the calibration instances. Computed by the :py:func:`fit` method.
            Must be *sorted* in increasing order.

    Examples:
        >>> train, test = next(LOOSampler(Table('housing')))
        >>> train, calibrate = next(RandomSampler(train, 2, 1))
        >>> icr = InductiveRegressor(AbsError(LinearRegressionLearner()), train, calibrate)
        >>> print(icr(test[0], 0.1))
    """

    def __init__(self, nc_measure, train=None, calibrate=None):
        """Initialize inductive regressor with a nonconformity measure, training set and calibration set.
        If present, fit the conformal regressor to the training set and compute the nonconformity scores of
        calibration set.

        Args:
            nc_measure (RegrNC): Regression nonconformity measure.
            train (Optional[Table]): Table of examples used as a training set.
            calibrate (Optional[Table]): Table of examples used as a calibration set.
        """
        super().__init__(nc_measure)
        if train is not None and calibrate is not None:
            self.fit(train, calibrate)

    def fit(self, train, calibrate):
        """Fit the conformal regressor to the training set, compute and store sorted nonconformity scores (:py:attr:`alpha`)
        on the calibration set and store the domain.

        Args:
            train (Optional[Table]): Table of examples used as a training set.
            calibrate (Optional[Table]): Table of examples used as a calibration set.
        """
        self.domain = train.domain
        self.nc_measure.fit(train)
        self.alpha = [self.nc_measure.nonconformity(inst) for inst in calibrate]
        self.alpha = np.array(sorted(self.alpha, reverse=True))


class CrossRegressor(InductiveRegressor):
    """Cross regression.

    Examples:
        >>> train, test = next(LOOSampler(Table('housing')))
        >>> ccr = CrossRegressor(AbsError(LinearRegressionLearner()), 4, train)
        >>> print(ccr(test[0], 0.1))
    """

    def __init__(self, nc_measure, k, train=None):
        """Initialize cross regressor with a nonconformity measure, number of folds and training set.
        If present, fit the conformal regressor to the training set.

        Args:
            nc_measure (RegrNC): Regression nonconformity measure.
            k (int): Number of folds.
            train (Optional[Table]): Table of examples used as a training set.
        """
        # store the unfitted nonconformity measure for making copies to fit on individual folds
        super().__init__(nc_measure)
        self.nc_measure_base = deepcopy(self.nc_measure)
        self.k = k
        if train is not None:
            self.fit(train)

    def fit(self, train):
        """Fit the cross regressor to the training set. Split the training set into k folds for use as
        training and calibration set with an inductive regressor. Concatenate the computed nonconformity scores
        and store them (:py:attr:`InductiveRegressor.alpha`).

        Args:
            train (Table): Table of examples used as a training set.
        """
        self.domain = train.domain
        self.nc_measure.fit(train)
        self.alpha = np.array([])
        for train_index, calibrate_index in KFold(self.k, shuffle=True).split(train):
            icr = InductiveRegressor(deepcopy(self.nc_measure_base), train[train_index], train[calibrate_index])
            self.alpha = np.concatenate((self.alpha, icr.alpha))
        self.alpha = np.array(sorted(self.alpha, reverse=True))


class LOORegressor(CrossRegressor):
    """Leave-one-out regressor is a cross conformal regressor with the number of folds equal
    to the size of the training set.

    Examples:
        >>> train, test = next(LOOSampler(Table('housing')))
        >>> ccr = LOORegressor(AbsError(LinearRegressionLearner()), train)
        >>> print(ccr(test[0], 0.1))
    """

    def __init__(self, nc_measure, train=None):
        super().__init__(nc_measure, 0, train)

    def fit(self, train):
        self.k = len(train)
        super().fit(train)



from unittest import TestCase

import Orange

import orangecontrib.conformal as cp


class TestSamplers(TestCase):
    def setUp(self):
        self.data = Orange.data.Table('iris')

    def test_random(self):
        a, b = 3, 2
        s = cp.evaluation.RandomSampler(self.data, a, b)
        train, test = next(s)
        self.assertTrue(isinstance(train[0], Orange.data.RowInstance))
        self.assertAlmostEqual(len(train)/len(test), a/b)

    def test_cross(self):
        folds = 7
        s = cp.evaluation.CrossSampler(self.data, k=folds)
        l = [(len(train), len(test)) for train, test in s]
        self.assertEqual(len(l), folds)
        self.assertTrue(all(a+b == len(self.data) for a, b in l))
        t = [b for a, b in l]
        self.assertLessEqual(max(t)-min(t), 1)

        s = cp.evaluation.CrossSampler(self.data, k=folds)
        ids = frozenset(frozenset(inst.id for inst in test) for train, test in s)
        self.assertEqual(len(ids), folds)

    def test_loo(self):
        s = cp.evaluation.LOOSampler(self.data)
        x = [len(test) == 1 for train, test in s]
        self.assertTrue(all(x))
        self.assertEqual(len(x), len(self.data))

    def test_repeat(self):
        rep = 5
        s = cp.evaluation.RandomSampler(self.data, 3, 2)
        ids = frozenset(frozenset(inst.id for inst in test) for train, test in s.repeat(5))
        self.assertEqual(len(ids), 5)

        s = cp.evaluation.CrossSampler(self.data, 3)
        ids = frozenset(frozenset(inst.id for inst in test) for train, test in s.repeat(5))
        self.assertEqual(len(ids), 15)


class TestEvaluation(TestCase):
    def test_run(self):
        tab = Orange.data.Table('iris')[::2]
        nc = cp.nonconformity.InverseProbability(Orange.classification.LogisticRegressionLearner())
        pred = cp.classification.InductiveClassifier(nc)
        r = cp.evaluation.run(pred, 0.1, cp.evaluation.CrossSampler(tab, 3), rep=2)
        self.assertEqual(len(r.preds), 2*len(tab))

        tab = Orange.data.Table('housing')[:100]
        nc = cp.nonconformity.AbsError(Orange.regression.LinearRegressionLearner())
        pred = cp.regression.InductiveRegressor(nc)
        r = cp.evaluation.run(pred, 0.1, cp.evaluation.RandomSampler(tab, 3, 1), rep=2)
        self.assertEqual(len(r.preds), 2*(1/4)*len(tab))

    def test_run_train_test(self):
        tab = Orange.data.Table('iris')
        nc = cp.nonconformity.InverseProbability(Orange.classification.LogisticRegressionLearner())
        pred = cp.classification.InductiveClassifier(nc)
        train, test, calibrate = tab[0::3], tab[1::3], tab[2::3]
        r = cp.evaluation.run_train_test(pred, 0.1, train, test, calibrate)
        self.assertEqual(len(r.preds), len(test))

    def test_accuracy(self):
        tab = Orange.data.Table('iris')
        nc = cp.nonconformity.InverseProbability(Orange.classification.LogisticRegressionLearner())
        pred = cp.classification.CrossClassifier(nc, 5)
        eps = 0.1
        res = cp.evaluation.run(pred, eps, cp.evaluation.CrossSampler(tab, 3))
        acc = res.accuracy()
        self.assertAlmostEqual(acc, 1-eps, delta=0.05)

    def test_time(self):
        tab = Orange.data.Table('iris')
        nc = cp.nonconformity.InverseProbability(Orange.classification.LogisticRegressionLearner())
        pred = cp.classification.InductiveClassifier(nc)
        res = cp.evaluation.run(pred, 0.1, cp.evaluation.CrossSampler(tab, 3))
        self.assertLess(res.time(), 10.0)

    def test_results_class(self):
        tab = Orange.data.Table('iris')
        nc = cp.nonconformity.InverseProbability(Orange.classification.LogisticRegressionLearner())
        pred = cp.classification.CrossClassifier(nc, 3)

        r = cp.evaluation.run(pred, 0.5, cp.evaluation.RandomSampler(tab, 2, 1))
        self.assertGreater(r.empty_criterion(), 0.0)
        r = cp.evaluation.run(pred, 0.1, cp.evaluation.RandomSampler(tab, 2, 1))
        self.assertGreater(r.singleton_criterion(), 0.8)
        r = cp.evaluation.run(pred, 0.01, cp.evaluation.RandomSampler(tab, 2, 1))
        self.assertGreater(r.multiple_criterion(), 0.1)

        r = cp.evaluation.run(pred, 0.1, cp.evaluation.RandomSampler(tab, 2, 1))
        self.assertGreater(r.singleton_correct(), 0.8)
        self.assertGreater(r.confidence(), 0.9)
        self.assertGreater(r.credibility(), 0.4)

        self.assertAlmostEqual(sum(r.accuracy(class_value=c) for c in tab.domain.class_var.values)/3,
                               r.accuracy(),
                               delta=1e-2)

        self.assertGreater(r.accuracy(eps=0.01), r.accuracy())
        self.assertGreater(r.accuracy(), r.accuracy(eps=0.2))

    def test_results_regr(self):
        tab = Orange.data.Table('housing')[:300]
        nc = cp.nonconformity.AbsErrorKNN(Orange.distance.Euclidean(), 10, average=True)
        pred = cp.regression.InductiveRegressor(nc)
        train, test = next(cp.evaluation.RandomSampler(tab, 4, 1))
        r1 = cp.evaluation.run_train_test(pred, 0.1, train, test)
        r5 = cp.evaluation.run_train_test(pred, 0.5, train, test)
        self.assertGreater(r1.median_range(), r5.median_range())
        self.assertGreater(r1.mean_range(), r5.mean_range())
        self.assertGreater(r1.interdecile_range(), r5.interdecile_range())
        self.assertGreater(r1.interdecile_mean(), r5.interdecile_mean())
        self.assertGreater(r1.std_dev(), r5.std_dev())

    def test_pickle(self):
        import pickle, os
        train, test = next(cp.evaluation.RandomSampler(Orange.data.Table('iris'), 2, 1))
        train, cal = next(cp.evaluation.RandomSampler(train, 2, 1))
        nc = cp.nonconformity.InverseProbability(Orange.classification.NaiveBayesLearner())
        ic = cp.classification.InductiveClassifier(nc)
        ic.fit(train, cal)
        pred1 = ic(test[0], 0.1)
        with open('temp.cp','wb') as f:
            pickle.dump(ic, f)
        with open('temp.cp','rb') as f:
            ic2 = pickle.load(f)
        os.remove('temp.cp')
        pred2 = ic2(test[0], 0.1)
        self.assertListEqual(pred1, pred2)



"""Evaluation module contains methods for evaluation of conformal predictors.

Function :py:func:`run` produces Results of an appropriate type by using a Sampler on a given data set
to split it into a training and testing set.

Structure:

- Sampler (sampling methods)
    - :py:class:`RandomSampler`
    - :py:class:`CrossSampler`
    - :py:class:`LOOSampler`

- Results (evaluation results)
    - :py:class:`ResultsClass`
    - :py:class:`ResultsRegr`

- Evaluation methods
    - :py:func:`run`
    - :py:func:`run_train_test`
"""

import numpy as np
import time
from sklearn.model_selection import KFold

from Orange.classification import LogisticRegressionLearner
from Orange.data import Table

from orangecontrib.conformal.classification import TransductiveClassifier, ConformalClassifier, CrossClassifier
from orangecontrib.conformal.nonconformity import InverseProbability
from orangecontrib.conformal.regression import TransductiveRegressor, InductiveRegressor, CrossRegressor


class Sampler:
    """Base class for various data sampling/splitting methods.

    Attributes:
        data (Table): Data set for sampling.
        n (int): Size of the data set.

    Examples:
        >>> s = CrossSampler(Table('iris'), 4)
        >>> for train, test in s.repeat(3):
        >>>     print(train)
    """

    def __init__(self, data):
        """Initialize the data set."""
        self.data = data
        self.n = len(data)

    def __iter__(self):
        return self

    def __next__(self):
        """Extending samplers should implement the __next__ method to return the selected
        and remaining part of the data.
        """
        raise NotImplementedError

    def repeat(self, rep=1):
        """Repeat sampling several times."""
        for r in range(rep):
            for train, test in self:
                yield train, test


class RandomSampler(Sampler):
    """Randomly samples a subset of data in proportion a:b.

    Attributes:
        k (float): Size of the selected subset.

    Examples:
        >>> s = RandomSampler(Table('iris'), 3, 2)
        >>> train, test = next(s)
    """

    def __init__(self, data, a, b):
        """Initialize the data set and the size of the desired selection."""
        super().__init__(data)
        self.k = a*self.n//(a+b)

    def __iter__(self):
        """Return a special iterator over a single split of data."""
        yield next(self)

    def __next__(self):
        """Splits the data based on a random permutation."""
        perm = np.random.permutation(self.n)
        train_ind, test_ind = perm[:self.k], perm[self.k:]
        return self.data[train_ind], self.data[test_ind]


class CrossSampler(Sampler):
    """Sample the data in :py:attr:`k` folds. Shuffle the data before determining the folds.

    Attributes:
        k (int): Number of folds.

    Examples:
        >>> s = CrossSampler(Table('iris'), 4)
        >>> for train, test in s:
        >>>     print(train)
    """
    def __init__(self, data, k):
        super().__init__(data)
        self.k = k
        self.kf = None

    def __next__(self):
        """Compute the next fold. Initializes a new k-fold split on each repetition of the entire
        sampling procedure.
        """
        if self.kf is None:
            self.kf = iter(KFold(self.k, shuffle=True).split(self.data))
        try:
            train_ind, test_ind = next(self.kf)
            return self.data[train_ind], self.data[test_ind]
        except StopIteration:
            self.kf = None
            raise StopIteration


class LOOSampler(CrossSampler):
    """Leave-One-Out sampler is a cross sampler with the number of folds equal to the size of the data set.

    Examples:
        >>> s = LOOSampler(Table('iris'))
        >>> for train, test in s:
        >>>     print(len(test))
    """

    def __init__(self, data):
        super().__init__(data, len(data))


class Results:
    """Contains results of an evaluation of a conformal predictor
    returned by the :py:func:`run` function.

    Examples:
        >>> cp = CrossClassifier(InverseProbability(LogisticRegressionLearner()), 5)
        >>> r = run(cp, 0.1, RandomSampler(Table('iris'), 2, 1))
        >>> print(r.accuracy())
    """

    def __init__(self):
        self.preds = []
        self.refs = []
        self.tm = 0

    def add(self, pred, ref):
        """Add a new predicted and corresponding reference value."""
        self.preds.append(pred)
        self.refs.append(ref)

    def concatenate(self, r):
        """Concatenate another set of results."""
        self.preds += r.preds
        self.refs += r.refs
        self.tm += r.tm

    def accuracy(self):
        """Compute the accuracy of the predictor averaging verdicts of individual predictions. This is the fraction
        of instances that contain the actual/reference class among the predicted ones for classification and the fraction of
        instances that contain the actual value within the predicted range for regression."""
        v = [p.verdict(r) for p, r in zip(self.preds, self.refs)]
        return np.mean(v)

    def time(self):
        return self.tm


class ResultsClass(Results):
    """Results of evaluating a conformal classifier. Provides classification specific efficiency measures.

    Examples:
        >>> cp = CrossClassifier(InverseProbability(LogisticRegressionLearner()), 5)
        >>> r = run(cp, 0.1, RandomSampler(Table('iris'), 2, 1))
        >>> print(r.singleton_criterion())
    """

    def accuracy(self, class_value=None, eps=None):
        """Compute accuracy for test instances with a given class value. If this parameter is not given,
        compute accuracy over all instances, regardless of their class."""
        v = [p.verdict(r, eps) for p, r in zip(self.preds, self.refs) if class_value is None or r == class_value]
        return np.mean(v)

    def confidence(self):
        """Average confidence of predictions."""
        return np.mean([pred.confidence() for pred in self.preds])

    def credibility(self):
        """Average credibility of predictions."""
        return np.mean([pred.credibility() for pred in self.preds])

    def confusion(self, actual, predicted):
        """Compute the number of singleton predictions of class `predicted` when the actual class is `actual`.

        Examples:
            Drawing a confusion matrix.

            >>> data = Table('iris')
            >>> cp = CrossClassifier(InverseProbability(LogisticRegressionLearner()), 3)
            >>> r = run(cp, 0.1, RandomSampler(data, 2, 1))
            >>> values = data.domain.class_var.values
            >>> form = '{: >20}'*(len(values)+1)
            >>> print(form.format('actual\\predicted', *values))
            >>> for a in values:
            >>>     c = [r.confusion(a, p) for p in values]
            >>>     print(('{: >20}'*(len(c)+1)).format(a, *c))
                actual\predicted         Iris-setosa     Iris-versicolor      Iris-virginica
                     Iris-setosa                  18                   0                   0
                 Iris-versicolor                   0                  14                   4
                  Iris-virginica                   0                   0                  12
        """
        return sum(pred.classes()[0] == predicted and ref == actual
                   for pred, ref in zip(self.preds, self.refs) if len(pred.classes()) == 1)

    def multiple_criterion(self):
        """Number of cases with multiple predicted classes."""
        c = [len(pred.classes()) > 1 for pred in self.preds]
        return np.mean(c)

    def singleton_criterion(self):
        """Number of cases with a single predicted class."""
        c = [len(pred.classes()) == 1 for pred in self.preds]
        return np.mean(c)

    def empty_criterion(self):
        """Number of cases with no predicted classes."""
        c = [len(pred.classes()) == 0 for pred in self.preds]
        return np.mean(c)

    def singleton_correct(self):
        """Fraction of singleton predictions that are correct."""
        c = [pred.verdict(ref) for pred, ref in zip(self.preds, self.refs) if len(pred.classes()) == 1]
        return np.mean(c)


class ResultsRegr(Results):
    """Results of evaluating a conformal regressor. Provides regression specific efficiency measures.

    Examples:
        >>> ir = InductiveRegressor(AbsErrorKNN(Euclidean(), 10, average=True))
        >>> r = run(ir, 0.1, RandomSampler(Table('housing'), 2, 1))
        >>> print(r.interdecile_range())
    """

    def widths(self):
        return [pred.width() for pred in self.preds]

    def median_range(self):
        """Median width of predicted ranges."""
        return np.median(self.widths())

    def mean_range(self):
        """Mean width of predicted ranges."""
        return np.mean(self.widths())

    def std_dev(self):
        """Standard deviation of widths of predicted ranges."""
        return np.std(self.widths())

    def interdecile_range(self):
        """Difference between the first and ninth decile of widths of predicted ranges."""
        w = self.widths()
        return np.percentile(w, 90) - np.percentile(w, 10)

    def interdecile_mean(self):
        """Mean width discarding the smallest and largest 10% of widths of predicted ranges."""
        w = self.widths()
        decile = int(0.1*len(w))
        return np.mean(w[decile:len(w)-decile])


def run(cp, eps, sampler, rep=1):
    """Run method is used to repeat an experiment one or more times with different splits of the dataset
    into a training and testing set. The splits are defined by the provided sampler. The conformal predictor
    itself might further split the testing set internally for its computations (e.g. inductive or cross predictors).

    Run the conformal predictor `cp` on the datasets defined by the provided sampler and number of repetitions
    and construct the results. Fit the conformal predictor on each training set returned by the sampler and
    evaluate it on the corresponding test set.
    Inductive conformal predictors use one third of the training set (random subset) for calibration.

    For more control over the exact datasets used for training, testing and calibration see :py:func:`run_train_test`.

    Returns:
        :py:class:`ResultsClass` or :py:class:`ResultsRegr`

    Examples:
        >>> cp = CrossClassifier(InverseProbability(LogisticRegressionLearner()), 5)
        >>> r = run(cp, 0.1, CrossSampler(Table('iris'), 4), rep=3)
        >>> print(r.accuracy(), r.empty_criterion())

    The above example uses a :py:class:`CrossSampler` to define training and testing datasets. Each fold is used as the test
    set and the rest as a training set. The entire process is repeated three times with different fold splits
    and results in 3*n predictions, where n is the size of the dataset.
    """

    classification = isinstance(cp, ConformalClassifier)
    results = ResultsClass() if classification else ResultsRegr()
    for train, test in sampler.repeat(rep):
        r = run_train_test(cp, eps, train, test)
        results.concatenate(r)
    return results


def run_train_test(cp, eps, train, test, calibrate=None):
    """Fits the conformal predictor `cp` on the training dataset and evaluates it on the testing set.
    Inductive conformal predictors use the provided calibration set or default to extracting one third
    of the training set (random subset) for calibration.

    Returns:
        :py:class:`ResultsClass` or :py:class:`ResultsRegr`

    Examples:
        >>> tab = Table('iris')
        >>> cp = CrossClassifier(InverseProbability(LogisticRegressionLearner()), 4)
        >>> r = run_train_test(cp, 0.1, tab[:100], tab[100:])
        >>> print(r.accuracy(), r.singleton_criterion())
    """

    classification = isinstance(cp, ConformalClassifier)
    results = ResultsClass() if classification else ResultsRegr()
    start = time.time()
    if isinstance(cp, TransductiveClassifier) or isinstance(cp, TransductiveRegressor) or \
            isinstance(cp, CrossClassifier) or isinstance(cp, CrossRegressor):
        cp.fit(train)
    else:  # Inductive predictor
        if calibrate is None:
            train, calibrate = next(RandomSampler(train, 2, 1))
        cp.fit(train, calibrate)
    for inst in test:
        results.add(cp.predict(inst, eps), inst.get_class())
    finish = time.time()
    results.tm = finish-start
    return results



#!/usr/bin/env python

from os import path, walk

import sys
from setuptools import setup, find_packages

NAME = "Orange3-Conformal"

VERSION = "1.1.0"

DESCRIPTION = "Orange3 Conformal Prediction library"
# pandoc --from=markdown --to=rst README.md -o README.rst
LONG_DESCRIPTION = open(path.join(path.dirname(__file__), 'README.rst')).read()
AUTHOR = 'Bioinformatics Laboratory, FRI UL'
AUTHOR_EMAIL = 'info@biolab.si'
URL = 'https://github.com/biolab/orange3-conformal'
LICENSE = 'GPLv3'

KEYWORDS = (
    # [PyPi](https://pypi.python.org) packages with keyword "orange3 add-on"
    # can be installed using the Orange Add-on Manager
	# uncomment when adding widgets
    #'orange3 add-on',
)

PACKAGES = find_packages()

PACKAGE_DATA = {
    'orangecontrib.conformal': ['datasets/*'],
}

DATA_FILES = [
    # Data files that will be installed outside site-packages folder
]

INSTALL_REQUIRES = [
    'Orange3',
]

ENTRY_POINTS = {
    # Entry points that marks this package as an orange add-on. If set, addon will
    # be shown in the add-ons manager even if not published on PyPi.
    'orange3.addon': (
        'conformal = orangecontrib.conformal',
    ),
    # Entry point used to specify packages containing tutorials accessible
    # from welcome screen. Tutorials are saved Orange Workflows (.ows files).
    #'orange.widgets.tutorials': (
    #    # Syntax: any_text = path.to.package.containing.tutorials
    #    'exampletutorials = orangecontrib.example.tutorials',
    #),

    # Entry point used to specify packages containing widgets.
    #'orange.widgets': (
    #    # Syntax: category name = path.to.package.containing.widgets
    #    # Widget category specification can be seen in
    #    #    orangecontrib/example/widgets/__init__.py
    #    'Examples = orangecontrib.example.widgets',
    #),

    # Register widget help
    #"orange.canvas.help": (
    #    'html-index = orangecontrib.example.widgets:WIDGET_HELP_PATH',)
}

NAMESPACE_PACKAGES = ["orangecontrib"]

TEST_SUITE = "orangecontrib.conformal.tests.suite"


def include_documentation(local_dir, install_dir):
    global DATA_FILES
    if 'bdist_wheel' in sys.argv and not path.exists(local_dir):
        print("Directory '{}' does not exist. "
              "Please build documentation before running bdist_wheel."
              .format(path.abspath(local_dir)))
        sys.exit(0)

    doc_files = []
    for dirpath, dirs, files in walk(local_dir):
        doc_files.append((dirpath.replace(local_dir, install_dir),
                          [path.join(dirpath, f) for f in files]))
    DATA_FILES.extend(doc_files)

if __name__ == '__main__':
    #include_documentation('doc/build/html', 'help/orange3-example')
    setup(
        name=NAME,
        version=VERSION,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        url=URL,
        license=LICENSE,
        packages=PACKAGES,
        package_data=PACKAGE_DATA,
        data_files=DATA_FILES,
        install_requires=INSTALL_REQUIRES,
        entry_points=ENTRY_POINTS,
        keywords=KEYWORDS,
        namespace_packages=NAMESPACE_PACKAGES,
        test_suite=TEST_SUITE,
        include_package_data=True,
        zip_safe=False,
    )



"""Nonconformity module contains nonconformity scores for classification and regression.

Structure:

- ClassNC (classification scores)
    - ClassModelNC (model based)
        :py:class:`InverseProbability`, :py:class:`ProbabilityMargin`, :py:class:`SVMDistance`,
        :py:class:`LOOClassNC`
    - ClassNearestNeighboursNC (nearest neighbours based)
        :py:class:`KNNDistance`, :py:class:`KNNFraction`
- RegrNC (regression scores)
    - RegrModelNC (model based)
        :py:class:`AbsError`, :py:class:`AbsErrorRF` :py:class:`AbsErrorNormalized`, :py:class:`LOORegrNC`,
        :py:class:`ErrorModelNC`
    - RegrNearestNeighboursNC (nearest neighbours based)
        :py:class:`AbsErrorKNN`, :py:class:`AvgErrorKNN`
"""

import math
from copy import deepcopy

import numpy as np

from Orange.base import Model
from Orange.data import Table, Instance
from Orange.distance import DistanceModel, Euclidean

from sklearn.ensemble import RandomForestRegressor
from sklearn.svm import SVC, LinearSVC, NuSVC


# ----- CLASSIFICATION ----- #

class ClassNC:
    """Base class for classification nonconformity scores.

    Extending classes should implement :py:func:`fit` and :py:func:`nonconformity` methods.
    """

    def __str__(self):
        return self.__class__.__name__

    def fit(self, data):
        """Process the data used for later calculation of nonconformities.

        Args:
            data (Table): Data set.
        """
        raise NotImplementedError

    def nonconformity(self, instance):
        """Compute the nonconformity score of the given `instance`."""
        raise NotImplementedError


# --- model-based --- #

class ClassModelNC(ClassNC):
    """Base class for classification nonconformity scores that are based on an underlying classifier.

    Extending classes should implement :py:func:`ClassNC.nonconformity` method.

    Attributes:
        learner: Untrained underlying classifier.
        model: Trained underlying classifier.
    """

    def __init__(self, classifier):
        """Store the provided classifier as :py:attr:`learner`."""
        self.learner = classifier
        self.model = None

    def __str__(self):
        return str(self.learner)

    def fit(self, data):
        """Train the underlying classifier on provided data and store the trained model."""
        self.model = self.learner(data)


class InverseProbability(ClassModelNC):
    """Inverse probability nonconformity score returns :math:`1 - p`, where :math:`p` is the probability
    assigned to the actual class by the underlying classification model (:py:attr:`ClassModelNC.model`).

    Examples:
        >>> train, test = next(LOOSampler(Table('iris')))
        >>> tp = TransductiveClassifier(InverseProbability(NaiveBayesLearner()), train)
        >>> print(tp(test[0].x, 0.1))
    """

    def __str__(self):
        return "InverseProbability ({})".format(super().__str__())

    def nonconformity(self, instance):
        predictions = self.model(instance, ret=Model.Probs)[0]
        return 1 - predictions[int(instance.get_class())]


class ProbabilityMargin(ClassModelNC):
    """Probability margin nonconformity score measures the difference :math:`d_p` between the predicted probability
    of the actual class and the largest probability corresponding to some other class. To put the values on scale
    from 0 to 1, the nonconformity function returns :math:`(1 - d_p) / 2`.

    Examples:
        >>> train, test = next(LOOSampler(Table('iris')))
        >>> tp = TransductiveClassifier(ProbabilityMargin(LogisticRegressionLearner()), train)
        >>> print(tp(test[0].x, 0.1))
    """

    def __str__(self):
        return "ProbabilityMargin ({})".format(super().__str__())

    def nonconformity(self, instance):
        predictions = self.model(instance, ret=Model.Probs)[0]
        y = int(instance.get_class())
        py = predictions[y]
        pz = max(p for z, p in enumerate(predictions) if z != y)
        return (1.0 - (py - pz)) / 2


class SVMDistance(ClassNC):
    """SVMDistance nonconformity score measures the distance from the SVM's decision boundary. The score depends
    on the distance and the side of the decision boundary that the example lies on.
    Examples that lie on the correct side of the decision boundary and would therefore result in a
    correct prediction using the SVM classifier have a nonconformity score less than 1, while the incorrectly
    predicted examples have a score more than 1.

    .. math::
        \\mathit{nc} =
        \\begin{cases}
        \\frac{1}{1+d} & \\text{correct}\\\\
        1+d &\\text{incorrect}
        \\end{cases}

    The provided SVM classifier must be a sklearn's SVM classifier (SVC, LinearSVC, NuSVC)
    providing the decision_function() which computes the distance to decision boundary. This nonconformity works
    only for binary classification problems.

    Examples:
        >>> from sklearn.svm import SVC
        >>> train, test = next(LOOSampler(Table('titanic')))
        >>> train, calibrate = next(RandomSampler(train, 2, 1))
        >>> icp = InductiveClassifier(SVMDistance(SVC()), train, calibrate)
        >>> print(icp(test[0].x, 0.1))
    """

    def __init__(self, classifier):
        assert isinstance(classifier, (SVC, LinearSVC, NuSVC)), \
            "Classifier must be a sklearn's SVM classifier (SVC, LinearSVC, NuSVC)."
        self.clf = classifier
        self.model = None

    def __str__(self):
        return "SVMDistance ({})".format(self.clf.__class__.__name__)

    def fit(self, data):
        assert len(data.domain.class_var.values) == 2, \
            "SVMDistance supports only binary classification problems."
        self.clf.fit(data.X, data.Y)

    def nonconformity(self, instance):
        y = self.clf.predict(np.atleast_2d(instance.x))[0]
        d = self.clf.decision_function(np.atleast_2d(instance.x))[0]
        if instance.y == y:
            return 1/(1+abs(d))
        else:
            return 1+abs(d)


# --- knn-based --- #

class NearestNeighbours:
    """Base class for nonconformity measures based on nearest neighbours.

    Attributes:
        distance: Distance measure.
        k (int):  Number of nearest neighbours.
    """

    def __init__(self, distance=Euclidean(), k=1):
        """Store the distance measure and the number of neighbours."""
        self.distance = distance
        self.k = k

    def fit(self, data):
        """Store the data for finding nearest neighbours."""
        self.data = data
        # fit the distance measure if uninitialized
        if not isinstance(self.distance, DistanceModel):
            self.distance = self.distance.fit(data)

    def neighbours(self, instance):
        """Compute distances to all other data instances using the distance measure (:py:attr:`distance`).

        Excludes data instances that are equal to the provided `instance`.

        Returns:
            List of pairs (distance, instance) in increasing order of distances.
        """
        other = self.data[np.array([not np.array_equal(row.x, instance.x) for row in self.data])]
        dist = self.distance(instance, other)[0]
        return sorted([(d, row) for d, row in zip(dist, other)], key=lambda x: x[0])


class ClassNearestNeighboursNC(NearestNeighbours, ClassNC):
    """Base class for nearest neighbrours based classification nonconformity scores."""
    pass


class KNNDistance(ClassNearestNeighboursNC):
    """Computes the sum of distances to k nearest neighbours of the same class as the given instance and
    the sum of distances to k nearest neighbours of other classes. Returns their ratio.

    Examples:
        >>> from Orange.distance import Euclidean
        >>> train, test = next(LOOSampler(Table('iris')))
        >>> cp = CrossClassifier(KNNDistance(Euclidean(), 10), 2, train)
        >>> print(cp(test[0].x, 0.1))
    """

    def __str__(self):
        return "KNNDistance (k={})".format(self.k)

    def nonconformity(self, instance):
        dist = self.neighbours(instance)
        same = [d for d, row in dist if row.get_class() == instance.get_class()]
        diff = [d for d, row in dist if row.get_class() != instance.get_class()]
        same, diff = same[:self.k], diff[:self.k]
        return sum(same) / sum(diff)


class KNNFraction(ClassNearestNeighboursNC):
    """Computes the k nearest neighbours of the given instance. Returns the fraction of instances of the
    same class as the given instance within its k nearest neighbours.

    Weighted version uses weights :math:`1/d_i` based on distances instead of simply counting the instances.
    Non-weighted version is equivalent to using a value 1 for all weights.

    Examples:
        >>> train, test = next(LOOSampler(Table('iris')))
        >>> cp = CrossClassifier(KNNFraction(Euclidean(), 10, weighted=True), 2, train)
        >>> print(cp(test[0].x, 0.1))
    """

    def __init__(self, distance=Euclidean(), k=1, weighted=False):
        super().__init__(distance, k)
        self.weighted = weighted

    def __str__(self):
        return "KNNFraction (k={})".format(self.k)

    def nonconformity(self, instance):
        dist = self.neighbours(instance)[:self.k]
        if self.weighted:
            diff = [1/d for d, row in dist if row.get_class() != instance.get_class()]
            return sum(diff) / sum(1/d for d, _ in dist)
        else:
            diff = [1 for d, row in dist if row.get_class() != instance.get_class()]
            return sum(diff) / len(dist)


# --- model-knn --- #

class LOOClassNC(NearestNeighbours, ClassNC):
    """
    .. math::
        \\mathit{nc} = \\mathit{error} + (1 - p)
        \\quad \\text{or} \\quad
        \\mathit{nc} = \\frac{1 - p}{\\mathit{error}}

    :math:`p` ... probability of actual class predicted from :math:`N_k(z^*)` - k nearest neighbours
    of the instance :math:`z^*`

    The first nonconformity score is used when the parameter :py:attr:`relative` is set to *False*
    and the second one when it is set to *True*.

    .. math::
        \\mathit{error} = \\frac {\\sum_{z_i \\in N_k(z^*)} w_i (1 - p_i)} {\\sum_{z_i \\in N_k(z^*)} w_i},
        \\quad w_i = \\frac{1}{d(x^*, x_i)}

    :math:`p_i` ... probability of actual class predicted from :math:`N_k(z') \\setminus z_i` or
    :math:`N_k(z') \\setminus z_i \\cup z^*` if the parameter :py:attr:`include` is set to *True*. :math:`z'` is
    :math:`z^*` if the :py:attr:`neighbourhood` parameter is '*fixed*' and :math:`z_i` if it's '*variable*'.
    """

    def __init__(self, classifier, distance=Euclidean(), k=10, relative=True, include=False, neighbourhood='fixed'):
        """Initialize the parameters."""
        super().__init__(distance, k)
        self.classifier = classifier
        self.relative = relative
        self.include = include
        assert neighbourhood in ['fixed', 'variable']
        self.neighbourhood = neighbourhood

    def fit(self, data):
        """Store the data for finding nearest neighbours and initialize cache."""
        super().fit(data)
        self.cache = {}

    def get_neighbourhood(self, inst):
        """Construct an Orange data Table consisting of instance's k nearest neighbours.
        Cache the results for later calls with the same instance."""
        t = tuple(inst.x)
        if t not in self.cache:
            neigh = self.neighbours(inst)[:self.k]
            neigh_d, neigh_list = zip(*neigh)
            neigh_tab = Table(neigh_list[0].domain, list(neigh_list))
            self.cache[t] = neigh_tab
        return self.cache[t]

    def error(self, inst, neighbours):
        """Compute the average weighted probability prediction error for predicting the actual class
        of each neighbour from the other ones. Include the new example among the neighbours
        if the parameter :py:attr:`include` is True."""
        sc = 0
        ws = []
        for i in range(len(neighbours)):
            neigh = Instance(neighbours.domain, neighbours[i])
            w = 1/self.distance(inst, neigh)
            ws.append(w)

            if self.neighbourhood == 'fixed':
                neighbours_i = neighbours[np.arange(len(neighbours)) != i]
            else:
                neighbours_i = self.get_neighbourhood(neigh)
            if self.include:
                neighbours_i = neighbours_i.copy()
                neighbours_i.append(inst)

            model = self.classifier(neighbours_i)
            sc += w * (1-model(neigh, ret=Model.Probs)[0][int(neigh.get_class())])
        return float(sc / sum(ws))

    def nonconformity(self, inst):
        neighbours = self.get_neighbourhood(inst)
        model = self.classifier(neighbours)
        error = self.error(inst, neighbours)
        yp = model(inst, ret=Model.Probs)[0][int(inst.get_class())]
        if self.relative:
            return (1-yp) / (1e-6 + error)
        else:
            return error + (1-yp)


# ----- REGRESSION ----- #

class RegrNC:
    """Base class for regression nonconformity scores.

    Extending classes should implement :py:func:`fit`, :py:func:`nonconformity` and :py:func:`predict` methods.
    """

    def __str__(self):
        return self.__class__.__name__

    def fit(self, data):
        """Process the data used for later calculation of nonconformities.

        Args:
            data (Table): Data set.
        """
        raise NotImplementedError

    def nonconformity(self, instance):
        """Compute the nonconformity score of the given `instance`."""
        raise NotImplementedError

    def predict(self, inst, nc):
        """Compute the inverse of the nonconformity score. Determine a range of values for which the
        nonconformity of the given `instance` does not exceed `nc`."""
        raise NotImplementedError


# --- model-based --- #

class RegrModelNC(RegrNC):
    """Base class for regression nonconformity scores that are based on an underlying classifier.

    Extending classes should implement :py:func:`RegrNC.nonconformity` and :py:func:`RegrNC.predict` methods.

    Attributes:
        learner: Untrained underlying classifier.
        model: Trained underlying classifier.
    """

    def __init__(self, classifier):
        """Store the provided classifier as :py:attr:`learner`."""
        self.learner = classifier
        self.model = None

    def __str__(self):
        return str(self.learner)

    def fit(self, data):
        """Train the underlying classifier on provided data and store the trained model."""
        self.model = self.learner(data)


class AbsError(RegrModelNC):
    """Absolute error nonconformity score returns the absolute difference between
    the predicted value (:math:`\\hat{y}`) by the underlying :py:attr:`RegrModelNC.model`
    and the actual value (:math:`y^{*}`).

    .. math::
        \\mathit{nc} = |\\hat{y}-y^{*}|

    Examples:
        >>> train, test = next(LOOSampler(Table('housing')))
        >>> cr = CrossRegressor(AbsError(LinearRegressionLearner()), 2, train)
        >>> print(cr(test[0].x, 0.1))
    """

    def __str__(self):
        return "AbsError ({})".format(super().__str__())

    def nonconformity(self, instance):
        return abs(instance.get_class() - float(self.model(instance)))

    def predict(self, inst, nc):
        y = float(self.model(inst))
        return y-nc, y+nc


class AbsErrorRF(RegrModelNC):
    """AbsErrorRF is based on an underlying regressor and a random forest. The prediction errors of regressor
    are used as nonconformity scores and are normalized by the standard deviation of predictions coming from
    individual trees in the forest.

    .. math::
        \\mathit{nc} = \\frac{|\\hat{y}-y^{*}|}{\sigma_\\mathit{RF} + \\beta}

    Examples:
        >>> from sklearn.ensemble import RandomForestRegressor
        >>> icr = InductiveRegressor(AbsErrorRF(RandomForestRegressionLearner(), RandomForestRegressor()))
        >>> r = run(icr, 0.1, CrossSampler(Table('housing'), 10))
        >>> print(r.accuracy(), r.median_range(), r.interdecile_mean())
    """

    def __init__(self, classifier, rf, beta=0.5):
        """Store the classifier and beta parameter."""
        assert isinstance(rf, RandomForestRegressor), \
            "Second parameter must be an instance of sklearn's RandomForestRegressor."
        self.learner, self.model = classifier, None
        self.rf = rf
        self.beta = beta

    def __str__(self):
        return "AbsErrorRF ({})".format(super().__str__())

    def fit(self, data):
        """Train the underlying classifier on provided data and store the trained model."""
        self.model = self.learner(data)
        self.rf = self.rf.fit(data.X, data.Y)

    def norm(self, inst):
        """Normalization factor is equal to the standard deviation of predictions from trees in a random forest
        plus a constant term :py:attr:`beta`."""
        ys = [estimator.predict(np.atleast_2d(inst.x))[0] for estimator in self.rf.estimators_]
        return np.std(ys) + self.beta

    def nonconformity(self, inst):
        y = float(self.model(inst))
        nc = abs(inst.get_class() - y)
        return nc / self.norm(inst)

    def predict(self, inst, nc):
        y = float(self.model(inst))
        norm = self.norm(inst)
        return y - nc*norm, y + nc*norm


class ErrorModelNC(RegrModelNC):
    """ErrorModelNC is based on two underlying regressors. The first one is trained to predict the value while the
    second one is used for predicting logarithms of the errors made by the first one.

    H. Papadopoulos and H. Haralambous. *Reliable prediction intervals with regression neural networks*.
    Neural Networks (2011).

    .. math::
        \\mathit{nc} = \\frac{|\\hat{y}-y^{*}|}{\\exp(\\mu)-1 + \\beta}

    :math:`\\mu` ... prediction for the value of :math:`\\log(|\\hat{y}-y^{*}|+1)` returned by the second regressor

    Parameter :py:attr:`loo` determines whether to use a leave-one-out schema for building the training set
    of errors for the second regressor or not.

    Examples:
        >>> nc = ErrorModelNC(SVRLearner(), LinearRegressionLearner())
        >>> icr = InductiveRegressor(nc)
        >>> r = run(icr, 0.1, CrossSampler(Table('housing'), 10))
        >>> print(r.accuracy(), r.median_range(), r.interdecile_mean())
    """

    def __init__(self, classifier, error_classifier, beta=0.5, loo=False):
        super().__init__(classifier)
        self.error_learner = error_classifier
        self.error_model = None
        self.beta = beta
        self.loo = loo

    def __str__(self):
        return "ErrorModelNC ({}, {})".format(self.learner, self.error_learner)

    def fit(self, data):
        if self.loo:
            ys = []
            for i in range(len(data)):
                train = data[np.arange(len(data)) != i]
                inst = data[i]
                self.model = self.learner(train)
                ys.append(math.log(abs(inst.get_class() - self.model(inst))+1))
            self.model = self.learner(data)
        else:
            self.model = self.learner(data)
            ys = [math.log(abs(row.get_class() - self.model(row))+1) for row in data]
            error_data = Table.from_numpy(data.domain, data.X, ys)
        error_data = Table.from_numpy(data.domain, data.X, ys)
        self.error_model = self.error_learner(error_data)

    def nonconformity(self, inst):
        nc = abs(inst.get_class() - float(self.model(inst)))
        norm = math.exp(float(self.error_model(inst)))-1 + self.beta
        return nc / norm

    def predict(self, inst, nc):
        y = float(self.model(inst))
        norm = math.exp(float(self.error_model(inst)))-1 + self.beta
        return y - nc*norm, y + nc*norm


class ExperimentalNC(RegrModelNC):
    def __init__(self, rf):
        self.rf = rf

    def fit(self, data):
        self.rf = self.rf.fit(data.X, data.Y)
        ys = sorted(data.Y)
        d = len(ys)//10
        self.ir = ys[-(d+1)] - ys[d]

    def norm(self, inst):
        ys = [estimator.predict(np.atleast_2d(inst.x))[0] for estimator in self.rf.estimators_]
        return np.std(ys) / self.ir

    def nonconformity(self, inst):
        y = float(self.rf.predict(np.atleast_2d(inst.x))[0])
        return abs(inst.get_class() - y / self.norm(inst))

    def predict(self, inst, nc):
        y = float(self.rf.predict(np.atleast_2d(inst.x))[0])
        norm = self.norm(inst)
        a, b = (y-nc)*norm, (y+nc)*norm
        return min(a,b), max(a,b)


# --- model-knn --- #

class AbsErrorNormalized(RegrModelNC, NearestNeighbours):
    """Normalized absolute error prediction uses an underlying regression model to predict the value,
    which is then normalized by the distance and variance of the nearest neighbours.

    H. Papadopoulos, V. Vovk and A. Gammerman. *Regression Conformal Prediction with Nearest Neighbours*.
    Journal of Artificial Intelligence Research (2011).

    .. math::
        \\mathit{nc} = \\frac{|\\hat{y}-y^{*}|}{\\exp(\\gamma \\lambda^*) + \exp(\\rho \\xi^*)}
        \\quad \\text{or} \\quad
        \\mathit{nc} = \\frac{|\\hat{y}-y^{*}|}{\\gamma + \\lambda^* + \\xi^*}

    The first nonconformity score is used when the parameter :py:attr:`exp` is set to *True*
    and the second one when it is set to *False*.

    .. math::
        \\lambda^* = \\frac{d_k(z^*)}{\\mathit{median}(\\{d_k(z), z \\in T\\})}, \\quad
        d_k(z) = \\sum_{z_i \\in N_k(z)} distance(x, x_i)

    .. math::
        \\xi^* = \\frac{\\sigma_k(z^*)}{\\mathit{median}(\\{\\sigma_k(z), z \\in T\\})}, \\quad
        \\sigma_k(z) = \\sqrt{\\frac{1}{k} \\sum_{z_i \\in N_k(z)}(y_i-\\bar{y})}

    Parameter :py:attr:`rf` enables the use of a random forest for computing the standard deviation
    of predictions instead of the nearest neighbours.
    """

    def __init__(self, classifier, distance=Euclidean(), k=10, gamma=0.5, rho=0.5, exp=True, rf=None):
        """Initialize the parameters."""
        RegrModelNC.__init__(self, classifier)
        NearestNeighbours.__init__(self, distance, k)
        self._gamma = gamma  # distance sensitivity
        self._rho = rho  # variance sensitivity
        self.exp = exp  # type of normalization
        self.rf = rf  # random forest for normalization
        if self.rf:
            assert isinstance(rf, RandomForestRegressor), \
                "Rf must be an instance of sklearn's RandomForestRegressor."

    def __str__(self):
        return "AbsErrorNormalized ({}, k={})".format(self.learner, self.k)

    def fit(self, data):
        """Train the underlying model and precompute medians for nonconformity scores."""
        RegrModelNC.fit(self, data)
        NearestNeighbours.fit(self, data)
        if self.rf:
            self.rf = self.rf.fit(data.X, data.Y)
        # compute medians in the training set used for normalization
        self.median_d = np.median([self._d(inst) for inst in self.data])
        self.median_s = np.median([self._sigma(inst) for inst in self.data])

    def _d(self, inst):
        """Sum of distances to nearest neighbours."""
        neigh = self.neighbours(inst)[:self.k]
        return sum(dist for dist, row in neigh)

    def _lambda(self, inst):
        """Normalized distance measure."""
        return self._d(inst) / self.median_d

    def _sigma(self, inst):
        """Standard deviation of y values. This comes either from the nearest neighbours or from the
        predictions of individual trees in a random forest if the :py:attr:`rf` is provided."""
        if not self.rf:
            neigh = self.neighbours(inst)[:self.k]
            return np.std([row.get_class() for dist, row in neigh])
        else:
            preds = [estimator.predict(np.atleast_2d(inst.x))[0] for estimator in self.rf.estimators_]
            return np.std(preds)

    def _xi(self, inst):
        """Normalized variance measure."""
        return self._sigma(inst) / self.median_s

    def norm(self, inst):
        """Compute the normalization factor."""
        if self.exp:
            return math.exp(self._gamma*self._lambda(inst)) + math.exp(self._rho*self._xi(inst))
        else:
            return self._gamma + self._lambda(inst) + self._xi(inst)

    def nonconformity(self, inst):
        yp = float(self.model(inst))
        return abs(inst.get_class() - yp) / self.norm(inst)

    def predict(self, inst, nc):
        y = float(self.model(inst))
        norm = self.norm(inst)
        return y-nc*norm, y+nc*norm


class LOORegrNC(NearestNeighbours, RegrNC):
    """
    .. math::
        \\mathit{nc} = \\mathit{error} + |\\hat{y}-y^{*}|
        \\quad \\text{or} \\quad
        \\mathit{nc} = \\frac{|\\hat{y}-y^{*}|}{\\mathit{error}}

    :math:`\\hat{y}` ... value predicted from :math:`N_k(z^*)`

    The first nonconformity score is used when the parameter :py:attr:`relative` is set to *False*
    and the second one when it is set to *True*.

    .. math::
        \\mathit{error} = \\frac {\\sum_{z_i \\in N_k(z^*)} w_i |\\hat{y_i}-y_i|} {\\sum_{z_i \\in N_k(z^*)} w_i},
        \\quad w_i = \\frac{1}{d(x^*, x_i)}

    :math:`\\hat{y_i}` ... value predicted from :math:`N_k(z') \\setminus z_i` or
    :math:`N_k(z') \\setminus z_i \\cup z^*` if the parameter :py:attr:`include` is set to *True*. :math:`z'` is
    :math:`z^*` if the :py:attr:`neighbourhood` parameter is '*fixed*' and :math:`z_i` if it's '*variable*'.
    """

    def __init__(self, classifier, distance=Euclidean(), k=10, relative=True, include=False, neighbourhood='fixed'):
        """Initialize the parameters."""
        super().__init__(distance, k)
        self.classifier = classifier
        self.relative = relative
        self.include = include
        assert neighbourhood in ['fixed', 'variable']
        self.neighbourhood = neighbourhood

    def fit(self, data):
        """Store the data for finding nearest neighbours and initialize cache."""
        super().fit(data)
        self.cache = {}

    def get_neighbourhood(self, inst):
        """Construct an Orange data Table consisting of instance's k nearest neighbours.
        Cache the results for later calls with the same instance."""
        t = tuple(inst.x)
        if t not in self.cache:
            neigh = self.neighbours(inst)[:self.k]
            neigh_d, neigh_list = zip(*neigh)
            neigh_tab = Table(neigh_list[0].domain, list(neigh_list))
            self.cache[t] = neigh_tab
        return self.cache[t]

    def error(self, inst, neighbours):
        """Compute the average weighted error for predicting the value of each neighbour from the other ones.
        Include the new example among the neighbours if the parameter :py:attr:`include` is True."""
        sc = 0
        ws = []
        for i in range(len(neighbours)):
            neigh = Instance(neighbours.domain, neighbours[i])
            w = 1/self.distance(inst, neigh)
            ws.append(w)

            if self.neighbourhood == 'fixed':
                neighbours_i = neighbours[np.arange(len(neighbours)) != i]
            else:
                neighbours_i = self.get_neighbourhood(neigh)
            if self.include:
                neighbours_i = neighbours_i.copy()
                neighbours_i.append(inst)

            model = self.classifier(neighbours_i)
            sc += w * abs(model(neigh) - neigh.get_class())
        return float(sc / sum(ws))

    def nonconformity(self, inst):
        neighbours = self.get_neighbourhood(inst)
        model = self.classifier(neighbours)
        error = self.error(inst, neighbours)
        yp = float(model(inst))
        if self.relative:
            return abs(yp - inst.get_class()) / error
        else:
            return error + abs(yp - inst.get_class())

    def predict(self, inst, nc):
        neighbours = self.get_neighbourhood(inst)
        model = self.classifier(neighbours)
        yp = float(model(inst))
        inst.set_class(yp)
        error = self.error(inst, neighbours)
        if self.relative:
            return yp - (nc*error), yp + (nc*error)
        else:
            return yp - (nc-error), yp + (nc-error)


# --- knn-based --- #

class RegrNearestNeighboursNC(NearestNeighbours, RegrNC):
    """Base class for nearest neighbours based regression nonconformity scores."""
    pass


class AbsErrorKNN(RegrNearestNeighboursNC):
    """Absolute error of k nearest neighbours computes the average value of the k nearest neighbours and
    returns an absolute difference between this average (:math:`y_k`) and the actual value (:math:`y^{*}`).

    .. math::
        \\bar{y} &= 1/k \sum_{N_k(x^{*})} y_i \\\\
        \\mathit{nc} &= |\\bar{y} - y^{*}|

    Weighted version can normalize by average and/or variance.

    .. math::
        \\mathit{nc} = \\frac{ |\\bar{y}-y^{*}| } { \\bar{y} \cdot y_{\\sigma} }

    Attributes:
        average (bool): Normalize by average.
        variance (bool): Normalize by variance.

    Examples:
        >>> train, test = next(LOOSampler(Table('housing')))
        >>> cr = CrossRegressor(AbsErrorKNN(Euclidean(), 10, average=True), 2, train)
        >>> print(cr(test[0].x, 0.1))
    """

    def __init__(self, distance=Euclidean(), k=10, average=False, variance=False):
        """Initialize the distance measure, number of nearest neighbours to consider and
        whether to normalize by average and by variance."""
        super().__init__(distance, k)
        self.average = average
        self.variance = variance

    def __str__(self):
        return "AbsErrorKNN (k={})".format(self.k)

    def stats(self, instance):
        """Computes mean and standard deviation of values within the k nearest neighbours."""
        dist = self.neighbours(instance)
        neigh = [row.get_class() for d, row in dist[:self.k]]
        return np.mean(neigh), np.std(neigh)

    def norm(self, avg, std):
        """Compute the normalization factor according to the chosen properties."""
        norm = 1
        if self.average: norm /= avg
        if self.variance: norm /= std
        return norm

    def nonconformity(self, instance):
        avg, std = self.stats(instance)
        nc = abs(instance.get_class() - avg)
        return nc * self.norm(avg, std)

    def predict(self, inst, nc):
        avg, std = self.stats(inst)
        norm = self.norm(avg, std)
        return avg - nc/norm, avg + nc/norm


class AvgErrorKNN(RegrNearestNeighboursNC):
    """Average error of k nearest neighbours computes the average absolute error of the actual value (:math:`y^{*}`)
    compared to the k nearest neighbours (:math:`y_i`).

    .. math::
        \\mathit{nc} = 1/k \sum_{N_k(x^{*})} |y^{*} - y_i|

    Note:
        There might be no suitable `y` values for the required significance level at the time of prediction.
        In such cases, the predicted range is [nan, nan].

    Examples:
        >>> train, test = next(LOOSampler(Table('housing')))
        >>> cr = CrossRegressor(AvgErrorKNN(Euclidean(), 10), 2, train)
        >>> print(cr(test[0].x, 0.1))
    """

    def __str__(self):
        return "AvgErrorKNN (k={})".format(self.k)

    def avg_abs(self, y, ys):
        return np.mean([abs(y-yi) for yi in ys])

    def avg_abs_inv(self, nc, ys):
        ys = sorted(ys)
        i, j = (len(ys)-1)//2, len(ys)//2
        if self.avg_abs(ys[i], ys) > nc:
            return np.nan, np.nan
        # lower bound
        while i-1 >= 0 and self.avg_abs(ys[i-1], ys) <= nc:
            i -= 1
        i1, i2 = i, len(ys)-i
        lo = ys[i] + len(ys)*(nc-self.avg_abs(ys[i], ys))/(i1-i2)
        # upper bound
        while j+1 < len(ys) and self.avg_abs(ys[j+1], ys) <= nc:
            j += 1
        j1, j2 = j+1, len(ys)-(j+1)
        hi = ys[j] + len(ys)*(nc-self.avg_abs(ys[j], ys))/(j1-j2)
        return lo, hi

    def nonconformity(self, instance):
        dist = self.neighbours(instance)[:self.k]
        ys = [row.get_class() for d, row in dist]
        return self.avg_abs(instance.get_class(), ys)

    def predict(self, inst, nc):
        dist = self.neighbours(inst)[:self.k]
        ys = [row.get_class() for d, row in dist]
        return self.avg_abs_inv(nc, ys)
